/*
 * Segment Datenübernahme und Aufbereitung (DUA), Fehleranalyse TLS spontan
 * Copyright (C) 2016 by Regierungspräsidium Tübingen, Landesstelle für Straßentechnik
 * Copyright 2016 by Kappich Systemberatung Aachen
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * Basiert auf der gleichnamigen Klasse aus de.bsvrz.dua.fehlertls.tls, BitCtrl Systems GmbH
 * 
 * SWE de.bsvrz.dua.fehlertlsspontan (Fehleranalyse TLS spontan)
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the 
 * GNU General Public License as published by the Free Software Foundation; either version 3 of 
 * the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 *
 * Dieses Programm ist freie Software. Sie können es unter den Bedingungen der GNU General Public 
 * License, wie von der Free Software Foundation veröffentlicht, weitergeben und/oder modifizieren, 
 * entweder gemäß Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, dass es Ihnen von Nutzen sein wird, 
 * aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder der 
 * VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License.
 * Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem Programm erhalten haben. 
 * Falls nicht, siehe <http://www.gnu.org/licenses/>.
 * Eine deutsche Übersetzung findet sich unter http://www.gnu.de/documents/gpl.de.html.
 *
 * Contact Information/Kontakt:
 * Regierungspräsidium Tübingen
 * Landesstelle für Straßentechnik
 * Heilbronner Straße 300 - 302
 * 70469 Stuttgart
 * Phone: +49 0711 8910-200
 * mailto: abteilung9@rpt.bwl.de
 */

package de.bsvrz.dua.fehlertlsspontan.tls;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dua.fehlertlsspontan.FehlerTlsSpontan;
import de.bsvrz.dua.fehlertlsspontan.enums.TlsDeFehlerStatus;
import de.bsvrz.dua.fehlertlsspontan.enums.TlsFehlerAnalyse;
import de.bsvrz.dua.fehlertlsspontan.parameter.IParameterTlsFehlerAnalyseListener;
import de.bsvrz.dua.fehlertlsspontan.parameter.ParameterTlsFehlerAnalyse;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageTemplate;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import de.bsvrz.sys.funclib.operatingMessage.OperatingMessage;

/**
 * TLS-Hierarchieelement DE.
 * Repräsentiert ein DE für die Fehleranalyse TLS spontan.
 * Meldet sich auf den globalen DE-Fehler an.
 * 
 * @author Regierungspräsidium Tübingen, Landesstelle für Straßentechnik, Thomas Bräuner
 */
public class De extends TlsGeraet implements ClientReceiverInterface, ClientSenderInterface, IParameterTlsFehlerAnalyseListener {
//
//	/**
//	 * Der zusätzliche Zeitverzug, der nach dem erwarteten Empfangszeitpunkt
//	 * noch bis zur Erkennung eines nicht gelieferten Messwertes abgewartet
//	 * werden muss.
//	 */
//	private long zeitVerzugFehlerErkennung = Long.MIN_VALUE;
//
//	/**
//	 * Der zusätzliche Zeitverzug, der nach der Fehlererkennung bis zur
//	 * Fehlerermittlung abgewartet werden muss.
//	 */
//	private long zeitVerzugFehlerErmittlung = Long.MIN_VALUE;

	/**
	 * <code>atg.tlsFehlerAnalyse</code>, <code>asp.analyse</code>.
	 */
	private static DataDescription fehlerDatenBeschreibung = null;
	
	/**
	 * <code>atg.tlsGloDeFehler</code>, <code>asp.tlsAntwort</code>.
	 */
	private static DataDescription nutzDatenBeschreibung = null;
	
	/**
	 * Fehlerstatus des DE; entspricht att.tlsDEFehlerStatus
	 * 0 = ok, 1 = StörEAK, 2 = StörSM
	 * außerdem -1 = keine Daten, also nicht initalisiert
	 */
	private int fehlerStatus = -1;
	
	/**
	 * Der DE hat eine Störung.
	 * Entweder, weil der Fehlerstatus StörEAK oder StörSM ist
	 * oder, weil der Datenzustand auf "Keine Daten" usw. gewechselt ist.
	 */
	private boolean stoerung = false;

	/**
	 * Der DE ist initalisiert, 
	 * d.h. er liefert Nutzdaten im Sinne der SWE.
	 */
	private boolean initialisiert = false;

	private static final MessageTemplate MESSAGE_TEMPLATE = new MessageTemplate(
			MessageGrade.WARNING,
			MessageType.APPLICATION_DOMAIN,
			MessageTemplate.object(),
			MessageTemplate.fixed(": "),
			MessageTemplate.variable("reason"),
			MessageTemplate.fixed(". "),
			MessageTemplate.ids()
	).withIdFactory(message -> message.getObject().getPidOrId() + " [DUA-FS-FU]");

	/**
	 * Standardkonstruktor.
	 * 
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param objekt
	 *            ein Systemobjekt vom Typ <code>typ.de</code>
	 * @param elter
	 *            das in der TLS-Hierarchie über diesem Gerät liegende Gerät
	 */
	protected De(ClientDavInterface dav, SystemObject objekt, TlsGeraet elter) {
		super(dav, objekt, elter);

		if (fehlerDatenBeschreibung == null) {
			fehlerDatenBeschreibung = new DataDescription(
					dav.getDataModel().getAttributeGroup("atg.tlsFehlerAnalyse"),
					dav.getDataModel().getAspect("asp.analyse"));
		}

		if (nutzDatenBeschreibung == null) {
			nutzDatenBeschreibung = new DataDescription(
					dav.getDataModel().getAttributeGroup("atg.tlsGloDeFehler"),
					dav.getDataModel().getAspect("asp.tlsAntwort"));
		}
		
		ParameterTlsFehlerAnalyse.getInstanz(dav,
				FehlerTlsSpontan.getTlsFehlerAnalyseObjekt()).addListener(this);
		
		dav.subscribeReceiver(this, objekt, nutzDatenBeschreibung, ReceiveOptions.normal(), ReceiverRole.receiver());
		Debug.getLogger().info("Überwache " + this.objekt.getPid() + ", " + nutzDatenBeschreibung);

		try {
			// Für DeWzg und DeVLT als Quelle für die Fehleranalyse anmelden.
			// Die anderen (zyklischen) DE werden von Fehleranlyse TLS verwaltet.
			if (objekt.isOfType("typ.deWzg") || objekt.isOfType("typ.deVlt")) {
				dav.subscribeSender(this, objekt, fehlerDatenBeschreibung, SenderRole.source());
			}
		} catch (OneSubscriptionPerSendData ignored) {
			// Wird ignoriert, da ein DE theoretisch in mehreren EAK vorkommen kann und dann hier versucht wird 2 Quellen anzumelden
		}
	}

	/**
	 * {@inheritDoc}
	 * <br>Hier Auswertung und Weiterverarbeitung des Fehlerstatus.
	 */
	@Override
	public void update(ResultData[] resultData) {
		for (ResultData data : resultData) {
			if (!data.hasData()) {
				// Keine Daten o.ä.
				if (fehlerStatus >= 0) {
					// Offensichtlich bricht gerade die Kommunikation weg!
					stoerung = true;
					initialisiert = false;
				}
				fehlerStatus = -1;
			} else {
				// Nutzdaten des Fehlerstatus empfangen
				if (data.getData() != null) {
					fehlerStatus = data.getData().getItem("DEFehlerStatus").asUnscaledValue().intValue();
					initialisiert = true;
					if (fehlerStatus > 0) {
						stoerung = true;
					} else {
						stoerung = false;
					}
				}
			}
//			System.out.println(fehlerStatus + " " + stoerung);
			// Betriebsmeldung senden
			if (objekt.isOfType("typ.deWzg") || objekt.isOfType("typ.deVlt")) {
				// Es ist ein DeWzg oder DeVlt
				if (stoerung) {
					// Es liegt eine Störung vor
					if (data.hasData()) {
						if (data.getData().getItem("DEKanalStatus").asUnscaledValue().intValue() == 0) {
							// Der DE ist nicht passiviert
							if (!((Eak)elter).alleDeGestoert(data.getDataTime())) {
								// Es sind nicht alle DE am Eltern-EAK gestört, sonst muss der EAK den Fehler publizieren
								String grund;
								if (fehlerStatus < 0) {
									grund = "DE ausgefallen. DE sendet keine Daten mehr.";
								} else {
									grund = "DE-Fehler " + TlsDeFehlerStatus.getZustand(fehlerStatus).toString() + ": Störung am ";
									grund += (fehlerStatus == 1) ? "EAK" : "SM";
									grund += " erkannt";
								}
								if (darfBetriebsmeldungSenden()) {
//									System.out.println("Betriebsmeldung wird gesendet: " + grund);
									publiziere(getMessage(grund));
								}
								publiziereFehler(data.getDataTime());
							}
						}
					}
				}
			}
		}
	}

	/**
	 * {@inheritDoc}
	 * <br>In diesem Fall ein DE.
	 */
	@Override
	public Art getGeraeteArt() {
		return Art.DE;
	}

	/**
	 * Publiziert eine erkannte Fehlerursache an diesem DE.
	 * 
	 * @param fehlerZeit
	 *            die Zeit mit der der Fehler assoziiert ist (Die Zeit, zu der
	 *            ausgefallene Datensatz erwartet wurde)
	 * @param tlsFehler
	 *            die Fehlerursache
	 */
	public final void publiziereFehlerUrsache(final long fehlerZeit,
			final TlsFehlerAnalyse tlsFehler) {
		Data datum = davVerbindung.createData(fehlerDatenBeschreibung.getAttributeGroup());
		datum.getUnscaledValue("TlsFehlerAnalyse").set(tlsFehler.getCode());
		try {
			davVerbindung.sendData(new ResultData(this.objekt, fehlerDatenBeschreibung,	fehlerZeit, datum));
		} catch (DataNotSubscribedException e) {
			Debug.getLogger().error("Datum " + datum + " konnte für " + this.objekt + 
					" nicht publiziert werden. Grund:\n" + e.getLocalizedMessage());
		} catch (SendSubscriptionNotConfirmed e) {
			Debug.getLogger().error("Datum " + datum + " konnte fuer " + this.objekt + 
					" nicht publiziert werden. Grund:\n" + e.getLocalizedMessage());
		}
	}

	/**
	 * {@inheritDoc}
	 * <br>In diesem Fall eine unbekannte Ursache.
	 */
	@Override
	public void publiziereFehler(long zeitStempel) {
		this.publiziereFehlerUrsache(zeitStempel, TlsFehlerAnalyse.UNBEKANNT);
	}

//	/**
//	 * {@inheritDoc}
//	 */
//	public synchronized void aktualisiereParameterTlsFehlerAnalyse(
//			long zeitverzugFehlerErkennung, long zeitverzugFehlerErmittlung) {
//		this.zeitVerzugFehlerErkennung = zeitverzugFehlerErkennung;
//		this.zeitVerzugFehlerErmittlung = zeitverzugFehlerErmittlung;
//	}

	private OperatingMessage getMessage(final String grund) {
		OperatingMessage message = MESSAGE_TEMPLATE.newMessage(objekt);
		message.put("reason", grund);
		message.addId("[DUA-FS-FU06]");
		return message;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
		// wird ignoriert (Anmeldung als Quelle)
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
		return false;
	}

	public boolean isGestoert() {
		return stoerung;
	}
	
	public boolean isInitialisiert() {
		return initialisiert;
	}
	
//	public boolean darfBetriebsmeldungSenden() {
//		if (zeitVerzugFehlerErkennung > 0 && zeitVerzugFehlerErmittlung > 0)
//			return true;
//		return false;
//	}
}
