/*
 * Segment Datenübernahme und Aufbereitung (DUA), Fehleranalyse TLS spontan
 * Copyright (C) 2016 by Regierungspräsidium Tübingen, Landesstelle für Straßentechnik
 * Copyright 2016 by Kappich Systemberatung Aachen
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * Basiert auf der gleichnamigen Klasse aus de.bsvrz.dua.fehlertls.tls, BitCtrl Systems GmbH
 * 
 * SWE de.bsvrz.dua.fehlertlsspontan (Fehleranalyse TLS spontan)
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the 
 * GNU General Public License as published by the Free Software Foundation; either version 3 of 
 * the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, see <http://www.gnu.org/licenses/>.
 *
 * Dieses Programm ist freie Software. Sie können es unter den Bedingungen der GNU General Public 
 * License, wie von der Free Software Foundation veröffentlicht, weitergeben und/oder modifizieren, 
 * entweder gemäß Version 3 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 * Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, dass es Ihnen von Nutzen sein wird, 
 * aber OHNE IRGENDEINE GARANTIE, sogar ohne die implizite Garantie der MARKTREIFE oder der 
 * VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License.
 * Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem Programm erhalten haben. 
 * Falls nicht, siehe <http://www.gnu.org/licenses/>.
 * Eine deutsche Übersetzung findet sich unter http://www.gnu.de/documents/gpl.de.html.
 *
 * Contact Information/Kontakt:
 * Regierungspräsidium Tübingen
 * Landesstelle für Straßentechnik
 * Heilbronner Straße 300 - 302
 * 70469 Stuttgart
 * Phone: +49 0711 8910-200
 * mailto: abteilung9@rpt.bwl.de
 */

package de.bsvrz.dua.fehlertlsspontan.tls;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dua.fehlertlsspontan.enums.TlsFehlerAnalyse;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageTemplate;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import de.bsvrz.sys.funclib.operatingMessage.OperatingMessage;

/**
 * TLS-Hierarchieelement Inselbus.
 * 
 * @author Regierungspräsidium Tübingen, Landesstelle für Straßentechnik, Thomas Bräuner
 */
public class Inselbus extends TlsGeraet {
	
	private final String port;

	private static final MessageTemplate TEMPLATE = new MessageTemplate(
			MessageGrade.ERROR,
			MessageType.APPLICATION_DOMAIN,
			MessageTemplate.fixed("Alle DE oder EAK am Inselbus "),
			MessageTemplate.variable("portIb"),
			MessageTemplate.fixed(" des "),
			MessageTemplate.object(),
			MessageTemplate.fixed(" defekt. DE und EAK instand setzen. "),
			MessageTemplate.ids()
	).withIdFactory(message -> message.getObject().getPidOrId() + " [DUA-FS-FU]");
	
//	private static final MessageTemplate TEMPLATE = new MessageTemplate(
//			MessageGrade.ERROR,
//			MessageType.APPLICATION_DOMAIN,
//			MessageTemplate.fixed("Alle DE oder EAK am Inselbus "),
//			MessageTemplate.object(),
//			MessageTemplate.fixed(" defekt. DE und EAK instand setzen. "),
//			MessageTemplate.ids()
//			).withIdFactory(message -> message.getObject().getPidOrId() + " [DUA-FS-FU]");
	
//	private static final MessageTemplate TEMPLATE_INSELBUS = new MessageTemplate(
//			MessageGrade.ERROR,
//			MessageType.APPLICATION_DOMAIN,
//			MessageTemplate.fixed("Inselbus "),
//			MessageTemplate.object(),
//			MessageTemplate.fixed(" gestört: Für die DE "),
//			MessageTemplate.set("sm", ", ", "des Steuermoduls ", "der Steuermodule "),
//			MessageTemplate.fixed(" sind keine Daten verfügbar. Inselbus "),
//			MessageTemplate.object(),
//			MessageTemplate.fixed(" instand setzen. "),
//			MessageTemplate.ids()
//	).withIdFactory(message -> message.getObject().getPidOrId() + " [DUA-FS-FU]");
	
	/**
	 * Standardkonstruktor.
	 * 
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param objekt
	 *            ein Systemobjekt vom Typ <code>typ.anschlussPunkt</code>
	 *            (unterhalb eines Objektes vom Typ <code>typ.kri</code>)
	 * @param elter
	 *            das in der TLS-Hierarchie ueber diesem Geraet liegende Geraet
	 */
	protected Inselbus(ClientDavInterface dav, SystemObject objekt,
			TlsGeraet elter) {
		super(dav, objekt, elter);
		this.port = "";

		/**
		 * Initialisiere Steuermodule
		 */
		for (SystemObject komPartner : this.objekt.getNonMutableSet("AnschlussPunkteKommunikationsPartner")
				.getElements()) {
			if (komPartner.isValid()) {
				Data konfigDatum = komPartner.getConfigurationData(dav.getDataModel()
						.getAttributeGroup("atg.anschlussPunktKommunikationsPartner"));
				if (konfigDatum != null) {
					SystemObject steuerModul = konfigDatum.getReferenceValue("KommunikationsPartner")
							.getSystemObject(); 
					if (steuerModul != null) {
						if (steuerModul.isOfType("typ.steuerModul")) {
							this.kinder.add(new Sm(dav, steuerModul, this));
						} else {
							Debug.getLogger().warning(
									"An " + komPartner + " (Inselbus: " + this.objekt
									+ ") dürfen nur Steuermodule definiert sein. Aber: "
									+ steuerModul + " (Typ: " + steuerModul.getType() + ")");
						}
					} else {
						Debug.getLogger().warning("An " + komPartner + " (Inselbus: " + this.objekt +
								") ist kein Steuermodul definiert.");
					}
				} else {
					Debug.getLogger()
							.warning("Konfiguration von " + komPartner + " (Inselbus: " + this.objekt +
									") konnte nicht ausgelesen werden. Das assoziierte Steuermodul wird ignoriert.");
				}
			}
		}

	}
	
	/**
	 * Konstruktor für den Inselbus.
	 * 
	 * @param dav
	 *            Datenverteiler-Verbindung
	 * @param objekt
	 *            ein Systemobjekt vom Typ <code>typ.anschlussPunkt</code>
	 *            (unterhalb eines Objektes vom Typ <code>typ.kri</code>)
	 *            als Platzhalter
	 * @param port
	 *            der Port des Anschlusspunkts, also der eigentliche Inselbus
	 * @param elter
	 *            das in der TLS-Hierarchie ueber diesem Geraet liegende Geraet
	 */
	protected Inselbus(ClientDavInterface dav, SystemObject objekt, String port, TlsGeraet elter) {
		super(dav, objekt, elter);
		this.port = port;
		
		/**
		 * Initialisiere Steuermodule
		 */
		final AttributeGroup atgAnschlussPunkt = dav.getDataModel().getAttributeGroup("atg.anschlussPunkt");
		final AttributeGroup atgAnschlussPunktKommunikationsPartner = dav.getDataModel().getAttributeGroup("atg.anschlussPunktKommunikationsPartner");
		for (final SystemObject apg : elter.objekt.getNonMutableSet("AnschlussPunkteGerät").getElements()) {
			if (apg.isValid()) {
				if (apg.getConfigurationData(atgAnschlussPunkt).getTextValue("PortNummer").getText().equals(port)) {
					for (final SystemObject apkp : ((ConfigurationObject)apg).
							getNonMutableSet("AnschlussPunkteKommunikationsPartner").getElements()) {
						if (apkp.isValid()) {
							Data konfigDatum = apkp.getConfigurationData(atgAnschlussPunktKommunikationsPartner);
							if (konfigDatum != null) {
								SystemObject steuerModul = konfigDatum.getReferenceValue("KommunikationsPartner")
										.getSystemObject(); 
								if (steuerModul != null) {
									if (steuerModul.isOfType("typ.steuerModul")) {
										this.kinder.add(new Sm(dav, steuerModul, this));
									} else {
										Debug.getLogger().warning(
												"An " + apkp + " (Inselbus: " + this.objekt
												+ ") dürfen nur Steuermodule definiert sein. Aber: "
												+ steuerModul + " (Typ: " + steuerModul.getType() + ")");
									}
								} else {
									Debug.getLogger().warning("An " + apkp + " (Inselbus: " + this.objekt +
											") ist kein Steuermodul definiert.");
								}
							} else {
								Debug.getLogger()
								.warning("Konfiguration von " + apkp + " (Inselbus: " + this.objekt +
										") konnte nicht ausgelesen werden. Das assoziierte Steuermodul wird ignoriert.");
							}
						}
					}
				}
			}
		}
	}

	/**
	 * {@inheritDoc}
	 * <br>In diesem Fall ein Inselbus.
	 */
	@Override
	public Art getGeraeteArt() {
		return Art.INSELBUS;
	}
	
	/**
	 * Gibt den Port also den Inselbus zurück.
	 * 
	 * @return den Inselbus
	 */
	public String getPort() {
		return port;
	}

	/**
	 * {@inheritDoc}
	 * In diesem Fall alle DE eines Inselbus defekt.
	 */
	@Override
	public void publiziereFehler(long zeitStempel) {
		OperatingMessage message = TEMPLATE.newMessage(elter.getObjekt());
		message.put("portIb", port);
		message.addId("[DUA-FS-FU02]");
		if (darfBetriebsmeldungSenden())
			this.publiziere(message);

		for (De de : getDes()) {
			if (de.getObjekt().isOfType("typ.deWzg") || de.getObjekt().isOfType("typ.deVlt")) {
				de.publiziereFehlerUrsache(zeitStempel,	TlsFehlerAnalyse.INSELBUS_ALLE_DE_DEFEKT);
			}
		}
	}
	
	/**
	 * Bestimmt, ob ein SM eine Betriebsmeldung absetzen darf
	 * oder ob das der Inselbus macht, weil alle DE auf Störung sind.
	 * 
	 * @param zeitStempel
	 *            der Zeitstempel des Fehlers
	 *            
	 * @return true, wenn alle DE gestört sind, sonst false
	 */
	public boolean alleDeGestoert(long zeitStempel) {
		boolean erg = true;
		for (De de : getDes()) {
			if (!de.isGestoert() && de.isInitialisiert()) {
				erg = false;
			}
		}
		if (erg) {
			if (!((Kri)elter).alleDeGestoert(zeitStempel)) {
				// Es sind nicht alle DE am Eltern-KRI gestört, sonst muss der KRI den Fehler publizieren
				publiziereFehler(zeitStempel);
			}
		}
		return erg;
	}

}
